@@ -236,9 +236,13 @@ DEBUGBITS(Debug_BIT)
  MSG(pmNoRegexMem,  etErr,  TRUE, 0,\
      "Unable to allocate regular expressions - no memory?\n")\
  MSG(pmRegexCompileFailed,  etWarn,  TRUE, 0,\
-     "Compilation of regular expression %s failed at offset %d with error %s.\n")\
+     "Compilation of regular expression %s failed with error %s.\n")\
  MSG(pmRegexMatchingError,  etErr,  TRUE, 0,\
-     "PCRE matching error %d.\n")
+     "PCRE matching error %d.\n") \
+ MSG(pmSuppTooHigh,  etWarn,  TRUE, 0,\
+     "Warning %d is numbered too high (max %d) and won't be suppressed.\n") \
+MSG(pmLongLines,  etWarn,  TRUE, 0,\
+     "ChkTeX does not handle lines over %d bytes correctly.  Some errors and line numbers may be wrong in this file.")
 
 #undef MSG
 #define MSG(num, type, inuse, ctxt, text) num,
@@ -299,12 +303,14 @@ extern FILE *OutputFile, *InputFile;
   DEF(int, AtLetter, FALSE) /* Whether `@' is a letter or not. */ \
   DEF(int, InHeader, TRUE)  /* Whether we're in the header */ \
   DEF(int, VerbMode, FALSE) /* Whether we're in complete ignore-mode */ \
-  DEF(long, MathMode, 0)     /* Whether we're in math mode or not */ \
-  DEF(const char *, VerbStr, "")   /* String we'll terminate verbmode upon */ \
+  DEF(long, MathMode, 0)    /* Whether we're in math mode or not */ \
+  DEF(const char *, VerbStr, "")     /* String we'll terminate verbmode upon */ \
   DEF(unsigned long, ErrPrint, 0)    /* # errors printed */ \
   DEF(unsigned long, WarnPrint, 0)   /* # warnings printed */ \
   DEF(unsigned long, UserSupp, 0)    /* # user suppressed warnings */ \
-  DEF(unsigned long, LineSupp, 0)    /* # warnings suppressed on a single line */
+  DEF(unsigned long, LineSupp, 0)    /* # warnings suppressed on a single line */ \
+  DEF(unsigned long long, FileSuppressions, 0)     /* # warnings suppressed in a file */ \
+  DEF(unsigned long long, UserFileSuppressions, 0) /* # User warnings suppressed in a file */
 
 #define DEF(type, name, value) extern type name;
 OPTION_DEFAULTS STATE_VARS
