@@ -33,14 +33,20 @@
 #include "Utility.h"
 #include "Resource.h"
 
+#if HAVE_PCRE || HAVE_POSIX_ERE
+
 #if HAVE_PCRE
+#include <pcreposix.h>
+#else
+#include <regex.h>
+#endif
 
-#include <pcre.h>
+#define REGEX_FLAGS REG_EXTENDED
+#define NUM_MATCHES 10
+#define ERROR_STRING_SIZE 100
 
-/* 3*(max_capturing_groups+1) */
-#define OVECCOUNT 3*(9+1)
-
-pcre** RegexArray = NULL;
+regex_t* RegexArray = NULL;
+regex_t* SilentRegex = NULL;
 int NumRegexes = 0;
 
 #endif
@@ -119,12 +125,25 @@ static const char LTX_SmallPunc[] = { '.', ',', 0 };
  * If more than one warning is to be suppressed, then multiple copies
  * of LineSuppDelim+number must be used.
  */
-const char *LineSuppDelim = "chktex ";
+const char LineSuppDelim[] = "chktex ";
+
+/*
+ * String used to delimit a file suppression.  This string must be
+ * followed immediately by the number of the warning to be suppressed.
+ * If more than one warning is to be suppressed, then multiple copies
+ * of FileSuppDelim+number must be used.
+ */
+const char FileSuppDelim[] = "chktex-file ";
 
 /*
  * A bit field used to hold the suppressions for the current line.
  */
 static unsigned long long LineSuppressions;
+/*
+ * A bit field used to hold the suppressions of numbered user warnings
+ * for the current line.
+ */
+static unsigned long long UserLineSuppressions;
 
 static unsigned long Line;
 
@@ -286,14 +305,14 @@ static char *MakeCpy(void)
 
 static char *PreProcess(void)
 {
-    /* Reset any line suppressions  */
-
-    LineSuppressions = 0LL;
-
-    /* Kill comments. */
-
     char *TmpPtr;
 
+    /* Reset any line suppressions  */
+
+    LineSuppressions = FileSuppressions;
+    UserLineSuppressions = UserFileSuppressions;
+
+    /* Kill comments. */
     strcpy(Buf, RealBuf);
 
     TmpPtr = Buf;
@@ -316,11 +335,57 @@ static char *PreProcess(void)
             /* Check for line suppressions */
             if (!NoLineSupp)
             {
-                ++TmpPtr;       /* move past NUL terminator */
-                while ((TmpPtr = strcasestr(TmpPtr, LineSuppDelim))) {
+                int error;
+                const int MaxSuppressionBits = sizeof(unsigned long long)*8-1;
+
+                /* Convert to lowercase to compare with LineSuppDelim */
+                EscapePtr = ++TmpPtr; /* move past NUL terminator */
+                while ( *EscapePtr )
+                {
+                    *EscapePtr = tolower(*EscapePtr);
+                    ++EscapePtr;
+                }
+
+                EscapePtr = TmpPtr; /* Save it for later */
+                while ((TmpPtr = strstr(TmpPtr, FileSuppDelim))) {
+                    TmpPtr += STRLEN(FileSuppDelim);
+                    error = atoi(TmpPtr);
+
+                    if (abs(error) > MaxSuppressionBits)
+                    {
+                        PrintPrgErr(pmSuppTooHigh, error, MaxSuppressionBits);
+                    }
+                    if (error > 0)
+                    {
+                        FileSuppressions |= (1LL << error);
+                        LineSuppressions |= (1LL << error);
+                    }
+                    else
+                    {
+                        UserFileSuppressions |= (1LL << (-error));
+                        UserLineSuppressions |= (1LL << (-error));
+                    }
+                }
+                TmpPtr = EscapePtr;
+
+                while ((TmpPtr = strstr(TmpPtr, LineSuppDelim))) {
+
                     TmpPtr += STRLEN(LineSuppDelim);
-                    int error = atoi(TmpPtr);
-                    LineSuppressions |= (1LL<<error);
+                    error = atoi(TmpPtr);
+
+                    if (abs(error) > MaxSuppressionBits)
+                    {
+                        PrintPrgErr(pmSuppTooHigh, error, MaxSuppressionBits);
+                    }
+
+                    if (error > 0)
+                    {
+                        LineSuppressions |= (1LL << error);
+                    }
+                    else
+                    {
+                        UserLineSuppressions |= (1LL << (-error));
+                    }
                 }
             }
             break;
@@ -467,11 +532,17 @@ static void WipeArgument(const char *Cmd, char *CmdPtr)
                 if (*TmpPtr == '[')
                     TmpPtr = GetLTXArg(TmpPtr, ArgBuffer, ']', NULL);
                 break;
+            case '(':
+                SKIP_AHEAD(TmpPtr, TmpC, LATEX_SPACE(TmpC));
+                if (*TmpPtr == '(')
+                    TmpPtr = GetLTXArg(TmpPtr, ArgBuffer, ')', NULL);
+                break;
             case '{':
                 SKIP_AHEAD(TmpPtr, TmpC, LATEX_SPACE(TmpC));
                 TmpPtr = GetLTXArg(TmpPtr, ArgBuffer, GET_TOKEN, NULL);
             case '}':
             case ']':
+            case ')':
                 break;
             default:
                 PrintPrgErr(pmWrongWipeTemp, &Cmd[strlen(Cmd) + 1]);
@@ -480,9 +551,9 @@ static void WipeArgument(const char *Cmd, char *CmdPtr)
         }
 
         if (TmpPtr)
-            strwrite(CmdPtr, VerbClear, TmpPtr - CmdPtr);
+            strwrite(CmdPtr+CmdLen, VerbClear, TmpPtr - CmdPtr - CmdLen);
         else
-            strxrep(CmdPtr, "()[]{}", *VerbClear);
+            strxrep(CmdPtr+CmdLen, "()[]{}", *VerbClear);
     }
 }
 
@@ -697,27 +768,32 @@ static void CheckRest(void)
 
     /* Search for user-specified warnings */
 
-#if ! HAVE_PCRE
+#if ! (HAVE_PCRE || HAVE_POSIX_ERE)
 
     if (INUSE(emUserWarnRegex) && UserWarnRegex.Stack.Used > 0)
     {
         PrintPrgErr(pmNoRegExp);
         ClearWord( &UserWarnRegex );
     }
+    else if (INUSE(emUserWarn))
+    {
+        strcpy(TmpBuffer, Buf);
+    }
 
 #else
 
     if (INUSE(emUserWarnRegex) && UserWarnRegex.Stack.Used > 0)
     {
-        const char *error;
-        int ErrOffset;
+        static char error[ERROR_STRING_SIZE];
+        static regmatch_t MatchVector[NUM_MATCHES];
+        int rc;
         int len = strlen(TmpBuffer);
-        static int ovector[OVECCOUNT];
+        strcpy(TmpBuffer, Buf);
 
         /* Compile all regular expressions if not already compiled. */
         if ( !RegexArray && UserWarnRegex.Stack.Used > 0 )
         {
-            RegexArray = (pcre**)malloc( sizeof(pcre*) * UserWarnRegex.Stack.Used );
+            RegexArray = (regex_t*)malloc( sizeof(regex_t) * UserWarnRegex.Stack.Used );
             if (!RegexArray)
             {
                 /* Allocation failed. */
@@ -731,29 +807,59 @@ static void CheckRest(void)
                 FORWL(Count, UserWarnRegex)
                 {
                     char *pattern = UserWarnRegex.Stack.Data[Count];
-                    RegexArray[NumRegexes] = pcre_compile( pattern, 0, &error, &ErrOffset, NULL);
+                    char *CommentEnd = NULL;
+
+                    /* See if it's got a special name that it goes by.
+                       Only use the comment if it's at the very beginning. */
+                    if ( strncmp(pattern,"(?#",3) == 0 )
+                    {
+                        CommentEnd = strchr(pattern, ')');
+                        /* TODO: check for PCRE/POSIX only regexes */
+                        *CommentEnd = '\0';
+                        /* We're leaking a little here, but this was never freed until exit anyway... */
+                        UserWarnRegex.Stack.Data[NumRegexes] = pattern+3;
+
+                        /* Compile past the end of the comment so that it works with POSIX too. */
+                        pattern = CommentEnd + 1;
+                    }
+
+                    /* Ignore PCRE and POSIX specific regexes.
+                     * This is mostly to make testing easier. */
+                    if ( strncmp(pattern,"PCRE:",5) == 0 )
+                    {
+                        #if HAVE_PCRE
+                        pattern += 5;
+                        #else
+                        continue;
+                        #endif
+                    }
+                    if ( strncmp(pattern,"POSIX:",6) == 0 )
+                    {
+                        #if HAVE_POSIX_ERE
+                        pattern += 6;
+                        #else
+                        continue;
+                        #endif
+                    }
+
+                    rc = regcomp((regex_t*)(&RegexArray[NumRegexes]),
+                                 pattern, REGEX_FLAGS);
 
                     /* Compilation failed: print the error message */
-                    if (RegexArray[NumRegexes] == NULL)
+                    if (rc != 0)
                     {
                         /* TODO: decide whether a non-compiling regex should completely stop, or just be ignored */
-                        PrintPrgErr(pmRegexCompileFailed, pattern, ErrOffset, error);
+                        regerror(rc,(regex_t*)(&RegexArray[NumRegexes]),
+                                 error, ERROR_STRING_SIZE);
+                        PrintPrgErr(pmRegexCompileFailed, pattern, error);
                     }
                     else
                     {
+                        if ( !CommentEnd )
+                        {
+                            ((char*)UserWarnRegex.Stack.Data[NumRegexes])[0] = '\0';
+                        }
                         ++NumRegexes;
-                        /* Only use the comment if it's at the very beginning. */
-                        if ( strncmp(pattern,"(?#",3) == 0 )
-                        {
-                            char *CommentEnd = strchr(pattern, ')');
-                            *CommentEnd = '\0';
-                            /* We're leaking a little here, but this was never freed until exit anyway... */
-                            UserWarnRegex.Stack.Data[Count] = pattern+3;
-                        }
-                        else
-                        {
-                            ((char*)UserWarnRegex.Stack.Data[Count])[0] = '\0';
-                        }
                     }
                 }
             }
@@ -762,53 +868,75 @@ static void CheckRest(void)
         for (Count = 0; Count < NumRegexes; ++Count)
         {
             int offset = 0;
+            char *ErrMessage = UserWarnRegex.Stack.Data[Count];
+            const int NamedWarning = strlen(ErrMessage) > 0;
+
             while (offset < len)
             {
+                /* Check if this warning should be suppressed. */
+                if (UserLineSuppressions != 0LL && NamedWarning)
+                {
+                    /* The warning can be named with positive or negative numbers. */
+                    int UserWarningNumber = abs(atoi(ErrMessage));
+                    if (UserLineSuppressions & (1LL << UserWarningNumber))
+                    {
+                        break;
+                    }
+                }
 
-                const int rc = pcre_exec( RegexArray[Count], NULL, TmpBuffer, len,
-                                          offset, 0, ovector, OVECCOUNT );
-
+                rc = regexec( (regex_t*)(&RegexArray[Count]), TmpBuffer+offset,
+                              NUM_MATCHES, MatchVector, 0);
                 /* Matching failed: handle error cases */
-                if (rc < 0)
+                if (rc != 0)
                 {
                     switch(rc)
                     {
-                        case PCRE_ERROR_NOMATCH:
+                        case REG_NOMATCH:
                             /* no match, no problem */
                             break;
                         default:
-                            PrintPrgErr(pmRegexMatchingError, rc);
+                            regerror(rc, (regex_t*)(&RegexArray[Count]),
+                                     error, ERROR_STRING_SIZE);
+                            PrintPrgErr(pmRegexMatchingError, error);
                             break;
                     }
+
                     offset = len; /* break out of loop */
                 }
                 else
                 {
-                    char *ErrMessage = UserWarnRegex.Stack.Data[Count];
-                    if ( strlen(ErrMessage) > 0 )
+#define MATCH (MatchVector[0])
+                    if ( NamedWarning )
                     {
                         /* User specified error message */
-                        PSERR2(ovector[0], ovector[1]-ovector[0], emUserWarnRegex,
+                        PSERR2(offset + MATCH.rm_so, MATCH.rm_eo - MATCH.rm_so,
+                               emUserWarnRegex,
                                strlen(ErrMessage), ErrMessage);
                     }
                     else
                     {
                         /* Default -- show the match */
-                        PSERR2(ovector[0], ovector[1]-ovector[0], emUserWarnRegex,
-                               ovector[1]-ovector[0], TmpBuffer+ovector[0]);
+                        PSERR2(offset + MATCH.rm_so, MATCH.rm_eo - MATCH.rm_so,
+                               emUserWarnRegex,
+                               MATCH.rm_eo - MATCH.rm_so,
+                               TmpBuffer + offset + MATCH.rm_so);
                     }
-                    offset = ovector[1];
+                    offset += MATCH.rm_eo;
+#undef MATCH
                 }
             }
         }
     }
+    else if (INUSE(emUserWarn))
+    {
+        strcpy(TmpBuffer, Buf);
+    }
 
 #endif
 
 
     if (INUSE(emUserWarn))
     {
-        strcpy(TmpBuffer, Buf);
         FORWL(Count, UserWarn)
         {
             for (UsrPtr = TmpBuffer;
@@ -916,7 +1044,13 @@ static void HandleBracket(int Char)
                         TmpPtr = BufPtr;
                         SKIP_AHEAD(TmpPtr, TmpC, TmpC == '}');
 
-                        if (!strchr(LTX_SmallPunc, *TmpPtr))
+                        /* If the next character is a period or comma,
+                         * or the last character is, then it's not an
+                         * error. */
+                        /* Checking 2 characters back seems dangerous,
+                         * but it's already done in CheckDash. */
+                        if ( !strchr(LTX_SmallPunc, *TmpPtr) &&
+                             !strchr(LTX_SmallPunc, *(TmpPtr-2)) )
                             HERE(1, emNoItFound);
                     }
 
@@ -990,6 +1124,95 @@ static void HandleBracket(int Char)
 }
 
 
+/*
+ * Checks to see if CmdBuffer matches any of the words in Silent, or
+ * any of the regular expressions in SilentCase.
+ *
+ */
+
+int CheckSilentRegex(void)
+{
+
+#if ! (HAVE_PCRE || HAVE_POSIX_ERE)
+
+    return HasWord(CmdBuffer, &Silent);
+
+#else
+
+    static char error[ERROR_STRING_SIZE];
+    char *pattern;
+    char *tmp;
+    int i;
+    int rc;
+    int len = 4;                /* Enough for the (?:) */
+
+    /* Initialize regular expression */
+    if (INUSE(emSpaceTerm) && SilentCase.Stack.Used > 0)
+    {
+        /* Find the total length we need */
+        /* There is 1 for | and the final for null terminator */
+        FORWL(i, SilentCase)
+        {
+            len += strlen( SilentCase.Stack.Data[i] ) + 1;
+        }
+
+        /* (A|B|...) */
+        tmp = (pattern = (char*)malloc( sizeof(char) * len ));
+
+        #if HAVE_PCRE
+        tmp = stpcpy(tmp,"(?:");
+        #else
+        tmp = stpcpy(tmp,"(");
+        #endif
+
+        FORWL(i, SilentCase)
+        {
+            tmp = stpcpy(tmp, SilentCase.Stack.Data[i]);
+            *tmp++ = '|';
+        }
+        tmp = stpcpy(--tmp, ")");
+
+        SilentRegex = malloc( sizeof(regex_t) );
+        rc = regcomp(SilentRegex, pattern, REGEX_FLAGS);
+
+        /* Compilation failed: print the error message */
+        if (rc != 0)
+        {
+            regerror(rc, SilentRegex, error, ERROR_STRING_SIZE);
+            PrintPrgErr(pmRegexCompileFailed, pattern, error);
+            SilentRegex = NULL;
+        }
+        /* Ensure we won't initialize it again */
+        SilentCase.Stack.Used = 0;
+        free(pattern);
+    }
+
+    /* Check against the normal */
+    if ( HasWord(CmdBuffer, &Silent) )
+        return 1;
+    if (!SilentRegex)
+        return 0;
+
+    /* Check against the regexes */
+    rc = regexec(SilentRegex, CmdBuffer, 0, NULL, 0);
+    if (rc == 0)
+        return 1;
+
+    /* Matching failed: handle error cases */
+    switch(rc)
+    {
+        case REG_NOMATCH:
+            return 0;
+            break;
+        default:
+            regerror(rc, SilentRegex, error, ERROR_STRING_SIZE);
+            PrintPrgErr(pmRegexMatchingError, error);
+            break;
+    }
+    return 0;
+
+#endif
+}
 
 /*
  * Searches the `Buf' for possible errors, and prints the errors. `Line'
@@ -1024,11 +1247,6 @@ int FindErr(const char *_RealBuf, const unsigned long _Line)
 
         BufPtr = SkipVerb();
 
-        if (!VerbMode)
-        {
-            CheckRest();
-        }
-
         while (BufPtr && *BufPtr)
         {
             PrePtr = BufPtr - 1;
@@ -1160,10 +1378,10 @@ int FindErr(const char *_RealBuf, const unsigned long _Line)
                     HEREA(3, emEllipsis, cTmpPtr);
                 }
 
-                /* Regexp: "([^A-Z@.])\.[.!?:;]*\s+[a-z]" */
+                /* Regexp: "([^A-Z@.])\.[.!?:]*\s+[a-z]" */
 
                 TmpPtr = BufPtr;
-                SKIP_AHEAD(TmpPtr, TmpC, strchr(LTX_GenPunc, TmpC));
+                SKIP_AHEAD(TmpPtr, TmpC, strchr(LTX_EosPunc, TmpC));
                 if (LATEX_SPACE(*TmpPtr))
                 {
                     if (!isupper((unsigned char)*PrePtr) && (*PrePtr != '@') &&
@@ -1347,7 +1565,7 @@ int FindErr(const char *_RealBuf, const unsigned long _Line)
                 }
 
                 if (LATEX_SPACE(*BufPtr) && !MathMode &&
-                    (!HasWord(CmdBuffer, &Silent)) &&
+                    !CheckSilentRegex() &&
                     (strlen(CmdBuffer) != 2))
                 {
                     PSERR(BufPtr - Buf, 1, emSpaceTerm);
@@ -1406,6 +1624,11 @@ int FindErr(const char *_RealBuf, const unsigned long _Line)
             }
         }
 
+        if (!VerbMode)
+        {
+            CheckRest();
+        }
+
     }
 
     return (TRUE);
